import math
import time


# Lifted from lumipy
class BackoffHandler:
    def __init__(
        self, pause_time: float = 1, max_pause_time: float = 10, beta: float = math.exp(1 / 200)
    ):
        if pause_time < 0.1 or pause_time > max_pause_time:
            raise ValueError(f"Pause time must be between 0.1 and {max_pause_time}, both inclusive.")
        if beta < 1:
            raise ValueError("Beta must be greater than 1.")

        self.max_pause_time = max_pause_time
        self.beta = beta
        self.pause_time = pause_time

    def _update_pause_time(self):
        self.pause_time = min(self.pause_time * self.beta, self.max_pause_time)

    def sleep(self):
        time.sleep(self.pause_time)
        self._update_pause_time()



from fbnconfig import Deployment
from fbnconfig.configuration import SystemConfigResource


# Not adding a test for it as if this runs in QA it may break pipeline tests due to system config changes
def example_configure(env):
    validate_instruments = SystemConfigResource(
        id="validate-instr",
        code="TransactionBooking",
        key="ValidateInstruments",
        value="True",
        description="Test from fbnconfig",
        default_value=False,
    )

    validate_txn_types = SystemConfigResource(
        id="validate-txn-types",
        code="TransactionBooking",
        key="ValidateTransactionTypes",
        value="False",
        description="Test from fbnconfig",
        default_value=True,
    )

    return Deployment("systemconfigtest", [validate_instruments, validate_txn_types])



from fbnconfig import Deployment, custom_data_model, datatype, property
from fbnconfig.coretypes import ResourceId


def configure(env):
    deployment_name = getattr(env, "name", "custom_data_model")

    # Create a property definition to use in the custom data model
    rating_prop = property.DefinitionResource(
        id="rating",
        domain=property.Domain.Instrument,
        scope=deployment_name,
        code="Rating",
        display_name="Credit Rating",
        data_type_id=datatype.DataTypeRef(id="default_str_rating", scope="system", code="string"),
        constraint_style=property.ConstraintStyle.Property,
        property_description="Credit rating for instruments",
        life_time=property.LifeTime.Perpetual,
    )

    maturity_prop = property.DefinitionResource(
        id="maturity",
        domain=property.Domain.Instrument,
        scope=deployment_name,
        code="MaturityDate",
        display_name="Maturity Date",
        data_type_id=datatype.DataTypeRef(id="default_str_maturity", scope="system", code="string"),
        constraint_style=property.ConstraintStyle.Property,
        property_description="Maturity date for instruments",
        life_time=property.LifeTime.TimeVariant,
    )

    # Create a base custom data model
    base_model = custom_data_model.CustomDataModelResource(
        id="base_instrument_model",
        entity_type="Instrument",
        resource_id=ResourceId(scope=deployment_name, code="BaseInstrumentModel"),
        display_name="Base Instrument Model",
        description="Base model for all instruments",
        conditions="InstrumentDefinition.InstrumentType eq 'Bond'",
        properties=[custom_data_model.DataModelProperty(property_key=rating_prop, required=True)],
    )

    # Create a child model that inherits from base
    bond_model = custom_data_model.CustomDataModelResource(
        id="bond_model",
        entity_type="Instrument",
        resource_id=ResourceId(scope=deployment_name, code="BondModel"),
        display_name="Bond Data Model",
        description="Validation model for bond instruments",
        parent_data_model=base_model,
        conditions="InstrumentDefinition.InstrumentType eq 'Bond'",
        properties=[custom_data_model.DataModelProperty(property_key=maturity_prop, required=True)],
        identifier_types=[
            custom_data_model.IdentifierType(identifier_key="Instrument/default/Isin", required=True)
        ],
        attribute_aliases=[
            custom_data_model.AttributeAlias(
                attribute_name=f"Properties[Instrument/{deployment_name}/MaturityDate]",
                attribute_alias="maturity",
            )
        ],
        recommended_sort_by=[
            custom_data_model.RecommendedSortBy(
                attribute_name=f"Properties[Instrument/{deployment_name}/MaturityDate]",
                sort_order=custom_data_model.SortOrder.ASC,
            )
        ],
    )

    return Deployment(deployment_name, [rating_prop, maturity_prop, base_model, bond_model])


if __name__ == "__main__":
    import os

    import click

    import fbnconfig

    @click.command()
    @click.argument("lusid_url", envvar="LUSID_ENV", type=str)
    @click.option("-v", "--vars_file", type=click.File("r"))
    def cli(lusid_url, vars_file):
        host_vars = fbnconfig.load_vars(vars_file)
        d = configure(host_vars)
        fbnconfig.deployex(d, lusid_url, os.environ["FBN_ACCESS_TOKEN"])

    cli()



from pydantic import BaseModel, Field, field_serializer, field_validator


class ComputedRead(BaseModel):
    scope: str = Field(init=True)
    code: str = Field(init=True)


class ComputedResource(ComputedRead):
    id: str = Field(exclude=True)


def test_validate():
    raw = {"scope": "sc1", "code": "cd1"}
    ext = {"id": raw["scope"] + "/" + raw["code"]}
    result = ComputedRead.model_validate(ext | raw)
    print(result.model_dump())
    assert result.model_dump() == {"scope": "sc1", "code": "cd1"}


class Reference():
    pass


class Child(BaseModel, Reference):
    id: str = Field(exclude=True)
    scope: str
    code: str


class Parent(BaseModel):
    id: str = Field(exclude=True)
    child: Child

    @field_serializer("child")
    def ser_child(self, v, info):
        if info.context and info.context.get("refs"):
            return "$ref:idc"
        return {"scope": v.scope, "code": v.code}

    @field_validator("child", mode="before")
    @classmethod
    def des_child(cls, v, info):
        print(info.context, "++++++++++++++++++++++++++")
        print(v, "++++++++++++++++++")
        if info.context and info.context.get("refs"):
            # return Child(id="idc", scope="ref", code="ref")
            return info.context["objects"][v["$ref"]]
        return v


def test_serialize_normal():
    ch = Child(id="idc", scope="sc1", code="cd1")
    pa = Parent(id="idp", child=ch)
    ser = pa.model_dump(context=None)
    assert ser == {"child": {"scope": "sc1", "code": "cd1"}}


def test_serialize_reference():
    ch = Child(id="idc", scope="sc1", code="cd1")
    pa = Parent(id="idp", child=ch)
    ser = pa.model_dump(context={"refs": True})
    assert ser == {"child": "$ref:idc"}


def test_deserialize_normal():
    ser = {"id": "idp", "child": {"id": "idc", "scope": "sc1", "code": "cd1"}}
    pa = Parent.model_validate(ser)
    assert pa.model_dump() == {"child": {"scope": "sc1", "code": "cd1"}}


def test_deserialize_reference():
    others = {"idc": Child(id="idc", scope="ref", code="ref")}
    ser = {"id": "idp", "child": {"$ref": "idc"}}
    pa = Parent.model_validate(ser, context={"refs": True, "objects": others})
    assert pa.model_dump() == {"child": {"scope": "ref", "code": "ref"}}



import os

import httpx


def raise_on_4xx_5xx(response):
    return response.raise_for_status()


def main():
    client = httpx.Client(
        base_url="https://fbn-qa.lusid.com",
        headers={"authorization": "Bearer " + os.getenv("FBN_ACCESS_TOKEN", "")},
        event_hooks={"response": [raise_on_4xx_5xx]},
    )
    try:
        client.get(url="/drive/api/folders/fb7225b61b864752aa03ba37872fe8e8")
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 404:
            print(e.response.reason_phrase)
            print("identifier already exists")
        print(e.response)


main()



import json
import string
from hashlib import sha256
from typing import Any, List

from pydantic import BaseModel, Field, model_validator

from fbnconfig import fund_accounting as fa
from fbnconfig.resource_abc import CamelAlias, FetchMixin, ListMixin, Resource, register_resource
from fbnconfig.urlfmt import Urlfmt

url: string.Formatter = Urlfmt("api/api/chartofaccounts")
CREATE_CLEARDOWN_MODULE_URL = ("{base}/{scope}/{code}/cleardownmodules")
CLEARDOWN_MODULE_URL = ("{base}/{scope}/{code}/cleardownmodules/{cleardown_module_code}")
CLEARDOWN_RULES_URL = ("{base}/{scope}/{code}/cleardownmodules/{cleardown_module_code}/cleardownrules")

GROUP = "Cleardown Configurations"


class CleardownModuleRule(CamelAlias, BaseModel):
    general_ledger_account_code: str
    rule_filter: str
    rule_id: str


@register_resource(
    group=GROUP,
    keys=["chartOfAccountsScope", "chartOfAccountsCode", "cleardownModuleCode"]
)
class CleardownModuleResource(Resource, BaseModel, ListMixin, FetchMixin):
    id: str = Field(exclude=True)
    chart_of_accounts: fa.ChartOfAccountsKey
    # Body
    cleardown_module_code: str
    display_name: str | None = None
    description: str | None = None
    rules: List[CleardownModuleRule] | None
    status: str = Field(default="Active", exclude=True)

    @property
    def scope(self):
        return self.chart_of_accounts.scope

    @property
    def code(self):
        return self.chart_of_accounts.code

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info):
        if not isinstance(data, dict):
            return data
        # Handle id from context (for dump/undump)
        if info.context and info.context.get("id"):
            data = data | {"id": info.context.get("id")}
        return data

    def read(self, client, old_state):
        scope, code = old_state.scope, old_state.code
        cleardown_module_code = old_state.cleardown_module_code
        formatted_url = url.format(
            CLEARDOWN_MODULE_URL,
            scope=scope,
            code=code,
            cleardown_module_code=cleardown_module_code
        )
        return client.get(formatted_url).json()

    def create(self, client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True,
                                  exclude={"scope", "code", "id", "chart_of_accounts"})
        cleardown_module_code = desired.pop("cleardownModuleCode", None)
        if cleardown_module_code is None:
            raise ValueError("cleardown_module_code must be provided")
        desired.update({"code": cleardown_module_code})

        sorted_desired = json.dumps(desired, sort_keys=True)
        content_hash = sha256(sorted_desired.encode()).hexdigest()
        formatted_url = url.format(CREATE_CLEARDOWN_MODULE_URL, scope=self.scope, code=self.code)
        client.request("POST", formatted_url, json=desired)

        # Serialize rules to dict for JSON storage
        rules_data = None
        if self.rules:
            rules_data = [rule.model_dump(mode="json", by_alias=True) for rule in self.rules]

        return {
            "scope": self.scope,
            "code": self.code,
            "cleardown_module_code": self.cleardown_module_code,
            "display_name": self.display_name,
            "description": self.description,
            "status": self.status,
            "rules": rules_data,
            "content_hash": content_hash
        }

    def update(self, client, old_state):
        if (
            (self.scope, self.code, self.cleardown_module_code) !=
            (old_state.scope, old_state.code, old_state.cleardown_module_code)
        ):
            self.delete(client, old_state)
            return self.create(client)

        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True,
                                  exclude={"scope", "code", "id", "chart_of_accounts"})

        cleardown_module_code = desired.pop("cleardownModuleCode", None)
        if cleardown_module_code is None:
            raise ValueError("cleardown_module_code must be provided")
        desired.update({"code": cleardown_module_code})

        sorted_desired = json.dumps(desired, sort_keys=True)
        desired_hash = sha256(sorted_desired.encode()).hexdigest()

        if desired_hash == old_state.content_hash:
            return None

        details_changed = (
            self.display_name != getattr(old_state, "display_name", None) or
            self.description != getattr(old_state, "description", None) or
            self.status != getattr(old_state, "status", None)
        )

        # Compare rules - old_state.rules are dicts, self.rules are CleardownModuleRule objects
        old_rules = getattr(old_state, "rules", None)
        current_rules_data = None
        if self.rules:
            current_rules_data = [rule.model_dump(mode="json", by_alias=True) for rule in self.rules]
        rules_changed = current_rules_data != old_rules

        if details_changed and rules_changed:
            self.update_details(client, desired_hash)
            return self.update_rules(client, desired_hash)

        if details_changed and not rules_changed:
            return self.update_details(client, desired_hash)

        if rules_changed and not details_changed:
            return self.update_rules(client, desired_hash)

        # If we get here, hash changed but we can't detect what changed
        # Fall back to recreating the resource
        self.delete(client, old_state)
        return self.create(client)

    @staticmethod
    def delete(client, old_state):
        scope, code = old_state.scope, old_state.code
        cleardown_module_code = old_state.cleardown_module_code
        formatted_url = url.format(
            CLEARDOWN_MODULE_URL,
            scope=scope,
            code=code,
            cleardown_module_code=cleardown_module_code
        )
        client.request("DELETE", formatted_url)

    def deps(self):
        deps = []

        # Always add chart of accounts dependency
        deps.append(self.chart_of_accounts)

        # Add account dependencies if rules exist
        if self.rules:
            # Get unique account codes from rules
            account_codes = set()
            for rule in self.rules:
                if rule.general_ledger_account_code:
                    account_codes.add(rule.general_ledger_account_code)

            # Create account references for each unique account code
            for account_code in account_codes:
                account_ref = fa.AccountRef(
                    id=f"account_{account_code}",
                    account_code=account_code,
                    chart_of_accounts=self.chart_of_accounts
                )
                deps.append(account_ref)

        return deps

    def update_details(self, client, desired_hash):
        metadata_json = {
            "displayName": self.display_name,
            "description": self.description,
            "status": self.status
        }
        try:
            client.request("PUT", url.format(
                CLEARDOWN_MODULE_URL,
                scope=self.scope,
                code=self.code,
                cleardown_module_code=self.cleardown_module_code
            ), json=metadata_json)
        except Exception as e:
            print(f"Error updating cleardown module metadata: {e}")
            raise e

        # Serialize rules to dict for JSON storage
        rules_data = None
        if self.rules:
            rules_data = [rule.model_dump(mode="json", by_alias=True) for rule in self.rules]

        return {
                "scope": self.scope,
                "code": self.code,
                "cleardown_module_code": self.cleardown_module_code,
                "display_name": self.display_name,
                "description": self.description,
                "status": self.status,
                "rules": rules_data,
                "content_hash": desired_hash
            }

    def update_rules(self, client, desired_hash):
        rules_data = []
        if self.rules:
            for rule in self.rules:
                rules_data.append(rule.model_dump(mode="json", by_alias=True))

        try:
            client.request("PUT", url.format(
                CLEARDOWN_RULES_URL,
                scope=self.scope,
                code=self.code,
                cleardown_module_code=self.cleardown_module_code
            ), json=rules_data)
        except Exception as e:
            print(f"Error updating cleardown rules: {e}")
            raise e

        # Serialize rules to dict for JSON storage
        rules_data = None
        if self.rules:
            rules_data = [rule.model_dump(mode="json", by_alias=True) for rule in self.rules]

        return {
                "scope": self.scope,
                "code": self.code,
                "cleardown_module_code": self.cleardown_module_code,
                "display_name": self.display_name,
                "description": self.description,
                "status": self.status,
                "rules": rules_data,
                "content_hash": desired_hash
            }

    @classmethod
    def list_endpoint(cls, **kwargs) -> str:
        return ("/api/api/chartofaccounts/{chartOfAccountsScope}/{chartOfAccountsCode}"
                "/cleardownmodules")

    @classmethod
    def fetch_endpoint(cls) -> str:
        return ("/api/api/chartofaccounts/{chartOfAccountsScope}/{chartOfAccountsCode}"
                "/cleardownmodules/{cleardownModuleCode}")

    @classmethod
    def create_resource_id(cls, data: dict) -> str:
        coa = data["chartOfAccountsId"]
        id_parts = f"{coa['scope']}_{coa['code']}_{data['cleardownModuleCode']}"
        return f"cleardown-module_{id_parts}".lower()

    @classmethod
    def get_keys(cls, data: dict[str, Any]) -> dict[str, Any]:
        keys = ("chartOfAccountsScope", "chartOfAccountsCode", "cleardownModuleCode")
        return {k: data[k] for k in keys}

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> dict[str, Any]:
        data.pop("href", None)
        data.pop("links", None)
        data.pop("version", None)
        data.pop("chartOfAccountsCode", None)
        data.pop("chartOfAccountsScope", None)

        data["id"] = cls.create_resource_id(data)

        chart_of_accounts_id = data.pop("chartOfAccountsId")
        chart_of_accounts_code = chart_of_accounts_id["code"]
        chart_of_accounts_scope = chart_of_accounts_id["scope"]
        data["chartOfAccounts"] = {
            "$ref": f"chart-of-account_{chart_of_accounts_scope}_{chart_of_accounts_code}".lower()
        }

        return data



import json

import pytest
from pydantic import BaseModel

from fbnconfig.resource_abc import CamelAlias


class InheritConfig(CamelAlias, BaseModel):
    snake_case_required: str
    snake_case_optional: int | None = None


@pytest.fixture()
def klass():
    return InheritConfig


def test_init_snake(klass):
    sut = klass(
        snake_case_required="a string"
    )
    assert sut.snake_case_required == "a string"


def test_init_camel(klass):
    # note: this is not quite the behaviour we wanted
    # but there does not seem to be a way to require snake for the init
    # and still be able to deserialize camel
    sut = klass(
        snakeCaseRequired="a string"
    )
    assert sut.snake_case_required == "a string"


def test_serialize(klass):
    sut = klass(
        snake_case_required="a string"
    )
    ser = sut.model_dump(mode="json", exclude_none=True, by_alias=True)
    assert ser == {"snakeCaseRequired": "a string"}


def test_deserialize_dict(klass):
    ser = {"snakeCaseRequired": "a string"}
    sut = klass.model_validate(ser)
    assert sut.snake_case_required == "a string"


def test_deserialize_json(klass):
    ser = json.dumps({"snakeCaseRequired": "a string"})
    sut = klass.model_validate_json(ser)
    assert sut.snake_case_required == "a string"



from fbnconfig import Deployment, access, drive, identity, scheduler

company = "robert.byrnemail.org"
service_user_login = "exports"
schedule_scope = "export_jobs"
drive_scope = "exports"
start_date = "2024-02-18T00:00:00.0000000+00:00"
role_name = "export_job_service"


def configure(host_vars):
    #
    # add a new service user
    #
    export_user = identity.UserResource(
        id="user2",
        first_name="Export",
        last_name="ServiceAccount",
        email_address=f"exports@{company}",
        login=f"exports@{company}",
        type=identity.UserType.SERVICE,
    )
    #
    # create role to allow the service user to be used as principal for the schedule
    # allow use of drive and luminesce
    #
    feature_selectors = [
        access.IdSelector(
            name="Execute * on honeycomb",
            description="Execute * on honeycomb",
            identifier={"scope": "Honeycomb", "code": "*"},
            actions=[access.ActionId(scope="Honeycomb", activity="Execute", entity="Feature")],
        ),
        access.IdSelector(
            name="Execute * on drive",
            description="Execute * on drive",
            identifier={"scope": "Drive", "code": "*"},
            actions=[access.ActionId(scope="LUSID", activity="Execute", entity="Feature")],
        ),
        access.IdSelector(
            name="Use drive on the web",
            identifier={"scope": "data-management", "code": "drive"},
            actions=[access.ActionId(scope="default", activity="view", entity="data-management")],
        ),
    ]
    feature_policy = access.PolicyResource(
        id="feature-policy",
        code="export_service_feature_policy",
        description="Grant access to honeycomb and drive features",
        applications=["Honeycomb", "Drive", "Website"],
        grant=access.Grant.ALLOW,
        selectors=feature_selectors,
        when=access.WhenSpec(activate=start_date),
    )
    # allow it to write to the target folder in drive
    data_selectors = [
        # allow read access on the root means you can see all the top level folders
        access.IdSelector(
            identifier={"path": "/"},
            actions=[access.ActionId(scope="default", activity="Read", entity="Folder")],
        ),
        # allow read access to the export folder itself
        access.IdSelector(
            identifier={"path": "/", "name": drive_scope},
            actions=[access.ActionId(scope="default", activity="Any", entity="Folder")],
        ),
        # allow creation of files inside the exports folder
        access.IdSelector(
            identifier={"name": "*", "path": f"/{drive_scope}"},
            actions=[access.ActionId(scope="default", activity="Any", entity="File")],
        ),
        # allow creation of folders inside the exports folder
        access.IdSelector(
            identifier={"name": "*", "path": f"/{drive_scope}"},
            actions=[access.ActionId(scope="default", activity="Any", entity="Folder")],
        ),
    ]
    data_policy = access.PolicyResource(
        id="data-policy",
        code="grant_export_data_folder",
        description="Grant access to exports folder",
        applications=["Drive"],
        grant=access.Grant.ALLOW,
        selectors=data_selectors,
        when=access.WhenSpec(activate=start_date),
    )
    # identity role
    identity_role = identity.IdentityRoleResource(
        id="identity_role", name=role_name, description="export_jobs"
    )
    # access role with name matching the identity role
    access_role = access.RoleResource(
        id="access_role",
        code=role_name,
        description="Grants access to the export folder",
        resource=access.RoleResourceRequest(
            policy_id_role_resource=access.PolicyIdRoleResource(policies=[feature_policy, data_policy])
        ),
        when=access.WhenSpec(activate=start_date),
        permission=access.Permission.READ,
    )
    # give the role to the user
    assign = identity.RoleAssignment(id="role_assignment", user=export_user, role=identity_role)
    #
    # create the folder
    #
    exports_folder = drive.FolderResource(id="exports_folder", name=drive_scope, parent=drive.root)
    #
    # create the schedule
    #
    # reference existing query runner job
    job = scheduler.JobRef(id="query_runner", scope="default", code="QueryRunner")
    # create a schedule for the job that is runAs the export user
    schedule = scheduler.ScheduleResource(
        id="portfolio_export_schedule",
        name="portfolio-export-schedule",
        scope=schedule_scope,
        code="portfolio_export",
        expression="0 0/5 * * * ?",
        timezone="Europe/London",
        job=job,
        description="Exports portfolios",
        use_as_auth=export_user,
        arguments={
            "QUERY_01": f"""
            @data = select distinct TableName from Sys.Field order by 1;
            @@ts = select strftime('%Y%m%d_%H%M%S');
            @x = use Drive.SaveAs with @data, @@ts
            --path=/{drive_scope}
            --type=CSV
            --fileNames=export1_{{@@ts}}.csv
            enduse;
            select * from @x
        """
        },
    )
    #
    # create the deployment
    #
    return Deployment(
        "scheduled_export",
        [export_user, identity_role, access_role, assign, exports_folder, job, schedule],
    )
